"""Aqueduct exceptions."""


class AqueductException(Exception):
    """Aqueduct Exception base class."""

    message: str

    def __init__(self, message: str):
        Exception.__init__(self, message)
        self.message = message


class FileNotFoundException(AqueductException):
    """File Not Found Exception."""

    def __init__(self, message: str):
        super().__init__(message)
        self.message = "File was not found on the server."


class InterruptedUploadException(AqueductException):
    """Interrupted Upload Exception."""

    def __init__(self, message: str):
        super().__init__(message)
        self.message = "Upload failed as upload was interrupted."


class InterruptedDownloadException(AqueductException):
    """Interrupted Download Exception"""

    def __init__(self, message: str):
        super().__init__(message)
        self.message = "Download failed as download was interrupted."



"""Experiment module."""

from __future__ import annotations

from datetime import datetime
from re import compile as recompile
from typing import List, Tuple
from uuid import UUID

from pydantic import BaseModel, Field

from pyaqueduct.client import AqueductClient

_MAX_TITLE_LENGTH = 100
_MAX_DESCRIPTION_LENGTH = 2000
_MAX_TAG_LENGTH = 50


def is_valid_tag(tag: str) -> bool:
    """Check if consists of only alphanumeric characters, underscore and hyphens"""
    pattern = recompile("^[a-zA-Z0-9_-]+$")

    return bool(pattern.match(tag))


class Experiment(BaseModel):
    """Experiment model."""

    client: AqueductClient
    "Client object reference."
    id: UUID
    "Unique ID of the experiment."
    alias: str
    "Alias of the experiment."
    created_at: datetime
    "Creation datetime of the experiment."

    @property
    def title(self) -> str:
        """Get title of experiment."""
        return self.client.get_experiment(experiment_uuid=str(self.id)).title

    @title.setter
    def title(self, value: str = Field(..., max_length=_MAX_TITLE_LENGTH)) -> None:
        """Set title of experiment.

        Args:
            value: New title.

        """
        self.client.update_experiment(experiment_uuid=str(self.id), title=value)

    @property
    def description(self) -> str:
        """Get description of experiment."""
        return self.client.get_experiment(str(self.id)).description

    @description.setter
    def description(self, value: str = Field(..., max_length=_MAX_DESCRIPTION_LENGTH)) -> None:
        """Set description of experiment.

        Args:
            value: New description.

        """
        self.client.update_experiment(experiment_uuid=str(self.id), description=value)

    @property
    def tags(self) -> List[str]:
        """Gets tags of experiment."""
        return self.client.get_experiment(str(self.id)).tags

    def add_tag(self, tag: str = Field(..., max_length=_MAX_TAG_LENGTH)) -> None:
        """Add new tag to experiment."""
        if not is_valid_tag(tag):
            raise ValueError(
                f"Tag {tag} should only contain alphanumeric characters, underscores or hyphens"
            )

        self.client.add_tag_to_experiment(experiment_uuid=str(self.id), tag=tag)

    def remove_tag(self, tag: str = Field(..., max_length=_MAX_TAG_LENGTH)) -> None:
        """Remove tag from experiment."""
        self.client.remove_tag_from_experiment(experiment_uuid=str(self.id), tag=tag)

    @property
    def files(self) -> List[Tuple[str, datetime]]:
        """Get file names of expriment."""
        return [
            (item.name, item.modified_at) for item in self.client.get_experiment(str(self.id)).files
        ]

    def download_file(self, file_name: str, destination_dir: str) -> None:
        """Download the specified file of experiment."""
        self.client.download_file(self.id, file_name=file_name, destination_dir=destination_dir)

    def upload_file(self, file: str) -> None:
        """Upload the specified file to experiment."""
        self.client.upload_file(self.id, file=file)

    @property
    def updated_at(self) -> datetime:
        """Get last updated datetime of the experiment."""
        return self.client.get_experiment(str(self.id)).updated_at



"""Aqueduct client module to communicate with the server instance."""

from pyaqueduct.client.client import AqueductClient
from pyaqueduct.client.types import ExperimentData, ExperimentFile, ExperimentsInfo

__all__ = ["AqueductClient", "ExperimentData", "ExperimentFile", "ExperimentsInfo"]



"""Aqueduct application programming interface (API) module."""

from __future__ import annotations

from typing import List

from pydantic import BaseModel, Field, HttpUrl, PositiveFloat, PrivateAttr

from pyaqueduct.client import AqueductClient
from pyaqueduct.experiment import _MAX_DESCRIPTION_LENGTH, _MAX_TITLE_LENGTH, Experiment


class API(BaseModel):
    """Aqueduct API interface to interact with experiments.

    Args:
        url: URL of the Aqueduct server including the prefix.
        timeout: timeout of operations in seconds.

    """

    url: HttpUrl
    timeout: PositiveFloat

    _client: AqueductClient = PrivateAttr()

    def __init__(self, url: str, timeout: float = 0.5):
        super().__init__(url=url, timeout=timeout)

        if not url.endswith("/"):
            url = url + "/"
        api_url = f"{url}api"

        self._client = AqueductClient(url=api_url, timeout=timeout)

    def create_experiment(
        self,
        title: str = Field(..., min_length=1, max_length=_MAX_TITLE_LENGTH),
        description: str = Field("", max_length=_MAX_DESCRIPTION_LENGTH),
    ) -> Experiment:
        """Create an experiment with specific title and description.

        Args:
            title: Title of the experiment.
            description: Description of the experiment.

        Returns:
            Experiment object to interact with its data.

        """

        experiment_data = self._client.create_experiment(title=title, description=description)
        return Experiment(
            client=self._client,
            id=experiment_data.id,
            alias=experiment_data.alias,
            created_at=experiment_data.created_at,
        )

    def get_experiment(self, alias: str) -> Experiment:
        """Get the experiment by the specified identifier to operate on.

        Args:
            experiment_id: ID of the specified experiment.

        Returns:
            Experiment object to interact with the experiment data.

        """
        experiment_data = self._client.get_experiment_by_alias(alias=alias)
        return Experiment(
            client=self._client,
            id=experiment_data.id,
            alias=experiment_data.alias,
            created_at=experiment_data.created_at,
        )

    def find_experiments(self, search: str, limit: int, offset: int = 0) -> List[Experiment]:
        """Find the experiments that have the search criteria in their title or id.

        Args:
            search: Search string to match.

        Returns:
            List of experiment objects to operate on their data.

        """
        experiments = self._client.get_experiments(
            title=search, limit=limit, offset=offset
        ).experiments
        return [
            Experiment(
                client=self._client,
                id=experiment.id,
                alias=experiment.alias,
                created_at=experiment.created_at,
            )
            for experiment in experiments
        ]



"""Aqueduct client class to enable experiment based operations."""

import logging
import os
from datetime import date
from http import HTTPStatus
from typing import List, Optional
from uuid import UUID

from gql import Client
from gql.client import SyncClientSession
from gql.transport.httpx import HTTPXTransport
from httpx import WriteTimeout, post, stream
from pydantic import BaseModel, HttpUrl, PrivateAttr
from tqdm import tqdm

from pyaqueduct.client.types import ExperimentData, ExperimentsInfo, TagsData
from pyaqueduct.exceptions import (
    InterruptedDownloadException,
    InterruptedUploadException,
)
from pyaqueduct.schemas.mutations import (
    add_tag_to_experiment_mutation,
    create_experiment_mutation,
    remove_tag_from_experiment_mutation,
    update_experiment_mutation,
)
from pyaqueduct.schemas.queries import (
    get_all_tags_query,
    get_experiment_query,
    get_experiments_query,
)


class AqueductClientResponse(BaseModel):
    """Response of methods of Aqueduct client"""

    result: str
    message: str


class AqueductClient(BaseModel):
    """
    AqueductClient - A client class for managing experiments, tags and files.

    This class provides methods for creating experiments, updating experiments,
    adding and removing tags, as well as uploading and downloading files.

    """

    url: HttpUrl
    timeout: float
    _gql_client: Client = PrivateAttr()
    _session: SyncClientSession = PrivateAttr()

    def __init__(self, url: str, timeout: float):
        """
        Args:
            url: URL of the Aqueduct server endpoint.
            timeout: Response timeout in seconds.

        """
        super().__init__(url=HttpUrl(url), timeout=timeout)

        self._gql_client = Client(transport=HTTPXTransport(url=f"{url}/graphql"))
        self.connect()

    def connect(self):
        """Connect to GraphQL connect"""
        self._session = self._gql_client.connect_sync()  # type: ignore

    def close(self):
        """Close connection with GraphQL client"""
        self._gql_client.close_sync()

    def create_experiment(
        self, title: str, description: str, tags: Optional[List[str]] = None
    ) -> ExperimentData:
        """
        Create an experiment with title, description and list of tags

        Args:
        - title (str): Title of experiment
        - description (str): Description of experiment
        - tags (List[str]): List of tags to be assigned to experiment

        Returns:
        Experiment: Experiment objects having all fields
        """

        experiment = self._session.execute(
            create_experiment_mutation,
            variable_values={"title": title, "description": description, "tags": tags or []},
        )

        experiment_obj = ExperimentData.from_dict(
            experiment["createExperiment"]  # pylint: disable=unsubscriptable-object
        )
        logging.info("Created experiment - %s - %s", experiment_obj.id, experiment_obj.title)
        return experiment_obj

    def update_experiment(
        self, experiment_uuid: str, title: Optional[str] = None, description: Optional[str] = None
    ) -> ExperimentData:
        """
        Update title or description or both for experiment

        Args:
        - experiment_uuid (UUID): UUID of experiment to be updated
        - title (str): New title of experiment
        - description (str): New description of experiment

        Returns:
        Experiment: Experiment object with updated fields
        """
        if title and title == "":
            raise ValueError("Title cannot be an empty string")

        if description and description == "":
            raise ValueError("Description cannot be an empty string")

        experiment = self._session.execute(
            update_experiment_mutation,
            variable_values={
                "experimentId": experiment_uuid,
                "title": title,
                "description": description,
            },
        )
        experiment_obj = ExperimentData.from_dict(
            experiment["updateExperiment"]  # pylint: disable=unsubscriptable-object
        )
        logging.info("Updated experiment - %s", experiment_obj.id)
        return experiment_obj

    def get_experiments(
        self,
        limit: int,
        offset: int,
        title: Optional[str] = None,
        tags: Optional[List[str]] = None,
        start_date: Optional[date] = None,
        end_date: Optional[date] = None,
    ) -> ExperimentsInfo:
        """
        Get a list of experiments

        Args:
        - limit (int): Number of experiments to be fetched
        - offset (offset): Number of experiments to skip
        - title (str): Perform search on experiments through title
        - tags (List[str]): Get experiments having these tags
        - start_date (date): Start date to filter experiments
        - end_date (date): End date to filter experiments to

        Returns:
        List[Experiment]: A list of experiments with filters applied
        """
        if limit <= 0:
            raise ValueError("Limit should be a positive number")

        experiments = self._session.execute(
            get_experiments_query,
            variable_values={
                "limit": limit,
                "offset": offset,
                "title": title,
                "start_date": start_date,
                "end_date": end_date,
                "tags": tags or [],
            },
        )

        experiments_obj = ExperimentsInfo.from_dict(
            experiments["experiments"]  # pylint: disable=unsubscriptable-object
        )
        logging.info(
            "Fetched %s experiments, total %s experiments",
            len(experiments_obj.experiments),
            experiments_obj.total_count,
        )
        return experiments_obj

    def get_experiment(self, experiment_uuid: str) -> ExperimentData:
        """
        Get an Experiment by ID or Alias

        Args:
        - id_ (str): "UUID | ALIAS" Unique identifier to be used
        - value (str): Value of unique identifier to use for fetching experiment

        Returns:
        Experiment: Experiment object having all fields
        """
        experiment = self._session.execute(
            get_experiment_query, variable_values={"type": "UUID", "value": experiment_uuid}
        )

        experiment_obj = ExperimentData.from_dict(
            experiment["experiment"]  # pylint: disable=unsubscriptable-object
        )
        logging.info("Fetched experiment - %s", experiment_obj.title)
        return experiment_obj

    def get_experiment_by_alias(self, alias: str) -> ExperimentData:
        """
        Get an Experiment by ID or Alias

        Args:
        - id_ (str): "UUID | ALIAS" Unique identifier to be used
        - value (str): Value of unique identifier to use for fetching experiment

        Returns:
        Experiment: Experiment object having all fields
        """
        experiment = self._session.execute(
            get_experiment_query, variable_values={"type": "ALIAS", "value": alias}
        )

        experiment_obj = ExperimentData.from_dict(
            experiment["experiment"]  # pylint: disable=unsubscriptable-object
        )
        logging.info("Fetched experiment - %s", experiment_obj.title)
        return experiment_obj

    def add_tag_to_experiment(self, experiment_uuid: str, tag: str) -> ExperimentData:
        """
        Add a tag to an experiment

        Args:
        - experiment_uuid (UUID): UUID of experiment to which tag is to be added
        - tag (str): Tag to be added to experiment

        Returns:
        Experiment: Experiment having tag added
        """

        experiment = self._session.execute(
            add_tag_to_experiment_mutation,
            variable_values={"experimentId": experiment_uuid, "tag": tag},
        )
        experiment_obj = ExperimentData.from_dict(
            experiment["addTagToExperiment"]  # pylint: disable=unsubscriptable-object
        )
        logging.info("Added tag %s to experiment <%s>", tag, experiment_obj.title)
        return experiment_obj

    def remove_tag_from_experiment(self, experiment_uuid: str, tag: str) -> ExperimentData:
        """
        Remove a tag from an experiment

        Args:
        - experiment_uuid (UUID): UUID of experiment frin which tag has to be removed
        - tag (str): Tag to be removed from experiment

        Returns:
        Experiment: Experiment having tag removed
        """

        experiment = self._session.execute(
            remove_tag_from_experiment_mutation,
            variable_values={"experimentId": experiment_uuid, "tag": tag},
        )

        experiment_obj = ExperimentData.from_dict(
            experiment["removeTagFromExperiment"]  # pylint: disable=unsubscriptable-object
        )
        logging.info("Removed tag %s from experiment <%s>", tag, experiment_obj.title)
        return experiment_obj

    def upload_file(self, experiment_uuid: UUID, file: str) -> AqueductClientResponse:
        """
        Upload file to a specific experiment.

        Args:
            experiment_uuid : The ID of the experiment.
            file_path: The local path to the file to be uploaded.

        Returns:
            Operation results object.

        """

        headers = {
            "file_name": os.path.basename(file),
        }

        upload_url = f"{self.url}/files/{str(experiment_uuid)}"
        with open(file, "rb") as files:
            try:
                resp = post(
                    upload_url, headers=headers, timeout=self.timeout, files={"file": files}
                )
            except InterruptedUploadException as exception:
                return AqueductClientResponse(result="failed", message=exception.message)

        if resp.status_code != 200:
            return AqueductClientResponse(
                result="failed", message=f"Upload failed {HTTPStatus(resp.status_code).name}"
            )
        logging.info("Successfully uploaded file {files}")
        return AqueductClientResponse(result="success", message="File uploaded successfully")

    def download_file(
        self, experiment_uuid: UUID, file_name: str, destination_dir: str
    ) -> AqueductClientResponse:
        """
        Download file from a specific experiment.

        Args:
            experiment_uuid: The ID of the experiment.
            file_id: The ID of the file to be downloaded.
            destination: The local path where the downloaded file will be saved.

        Returns:
            Operation results object.

        """
        download_url: str = f"{self.url}/files/{experiment_uuid}/{file_name}"
        destination = f"{destination_dir}/{file_name}"

        try:
            with open(destination, "wb") as download_file:
                with stream("GET", download_url, timeout=self.timeout) as response:
                    total = int(response.headers["Content-Length"])
                    with tqdm(
                        total=total, unit_scale=True, unit_divisor=1024, unit="B"
                    ) as progress:
                        num_bytes_downloaded = response.num_bytes_downloaded
                        for chunk in response.iter_bytes():
                            download_file.write(chunk)
                            progress.update(response.num_bytes_downloaded - num_bytes_downloaded)
                            num_bytes_downloaded = response.num_bytes_downloaded

        except InterruptedDownloadException as exception:
            return AqueductClientResponse(result="failed", message=exception.message)

        except WriteTimeout as exception:
            return AqueductClientResponse(
                result="failed", message=f"Timeout error for file download {exception}"
            )

        return AqueductClientResponse(result="success", message=f"File downloaded to {destination}")

    def get_tags(self, limit: int, offset: int, dangling: bool = True) -> TagsData:
        """
        Get a list of existing tags

        Args:
        - limit (int): Number of tags to be fetched
        - offset (offset): Number of tags to skip
        - dangling (bool): If tags not linked to any experiment should be included or not

        Returns:
        List[str]: A list of all existing tags
        """
        if limit <= 0:
            raise ValueError("Limit cannot be 0")

        tags = self._session.execute(
            get_all_tags_query,
            variable_values={"limit": limit, "offset": offset, "dangling": dangling},
        )

        tags_obj = TagsData.from_dict(tags["tags"])  # pylint: disable=unsubscriptable-object
        logging.info("Fetched %s tags, total %s tags", len(tags_obj.tags), tags_obj.total_count)
        return tags_obj



"""Aqueduct GraphQL Mutation schemas"""
from gql import gql

create_experiment_mutation = gql(
    """
    mutation CreateExperiment (
        $title: String!,
        $description: String!,
        $tags: [String!]!
    ) {
        createExperiment (
            createExperimentInput: {
                title: $title,
                description: $description,
                tags: $tags
            }
        ) {
            id
            title
            description
            alias
            createdAt
            updatedAt
            tags
            files {
                name
                path
                modifiedAt
            }
        }
    }
    """
)

update_experiment_mutation = gql(
    """
    mutation UpdateExperiment(
        $experimentId: UUID!,
        $title: String,
        $description: String
    ) {
        updateExperiment (
            experimentId: $experimentId,
            experimentUpdateInput: {
                title: $title,
                description: $description
            }
        ) {
            id
            title
            description
            alias
            createdAt
            updatedAt
            tags
            files {
                name
                path
                modifiedAt
            }
        }
    }
    """
)

add_tag_to_experiment_mutation = gql(
    """
    mutation AddTagToExperiment (
        $experimentId: UUID!,
        $tag: String!
    ) {
        addTagToExperiment (
            experimentTagInput: {
                experimentId: $experimentId,
                tag: $tag
            }
        ) {
            id
            title
            description
            alias
            createdAt
            updatedAt
            tags
            files {
                name
                path
                modifiedAt
            }
        }
    }
    """
)

remove_tag_from_experiment_mutation = gql(
    """
    mutation RemoveTagFromExperiment (
        $experimentId: UUID!,
        $tag: String!
    ) {
        removeTagFromExperiment (
            experimentTagInput: {
                experimentId: $experimentId,
                tag: $tag
            }
        ) {
            id
            title
            description
            alias
            createdAt
            updatedAt
            tags
            files {
                name
                path
                modifiedAt
            }
        }
    }
    """
)
